import boto3
import csv
import os
import json
import pandas as pd

# Function to find all S3 paths for the given RSN in the CSV file
def find_s3_paths_from_csv(rsn_number, csv_file):
    s3_paths = []
    with open(csv_file, mode='r') as file:
        csv_reader = csv.reader(file)
        next(csv_reader)  # Skip header row
        for row in csv_reader:
            csv_rsn_number, s3_path = row[0], row[1]
            if csv_rsn_number == rsn_number:
                s3_paths.append(s3_path)
    print(s3_paths)
    return s3_paths


# Function to download the file from S3 using the S3 path
def download_files_from_s3(s3_paths, download_directory):
    for s3_path in s3_paths:
        s3_key = s3_path.split(f"s3://{bucket_name}/")[-1]  # Extract the S3 key (path within the bucket)
        file_name = s3_key.split('/')[-1]  # Extract file name from the S3 key
        local_file_path = os.path.join(download_directory, file_name)

        # Download the file from S3
        s3_client.download_file(bucket_name, s3_key, local_file_path)
        print(f"File downloaded: {local_file_path}")


# Function to process the RSN list and download all matching files
def process_rsn_list(rsn_list, csv_file, download_directory):
    for rsn_number in rsn_list:
        # Find all matching S3 paths for the RSN number in the CSV file
        s3_paths = find_s3_paths_from_csv(rsn_number, csv_file)

        if s3_paths:
            print(f"Found {len(s3_paths)} files for RSN {rsn_number}. Downloading...")
            # Download all files for the RSN number
            download_files_from_s3(s3_paths, download_directory)
        else:
            print(f"No files found for RSN {rsn_number} in the CSV file.")

def convert_to_rsn_list(input_string):
    # Strip trailing commas and whitespace, then split by comma
    numbers = input_string.strip(',').split(',')
    # Prefix each number with 'RSN' and create a list
    rsn_list = [f'RSN{num}' for num in numbers if num.strip()]  # This also ensures no empty strings are processed
    return rsn_list

if __name__ == "__main__":
    s3_client = boto3.client(
        's3',
        aws_access_key_id='AKIAYS2NQF53LPIXAPGQ',
        aws_secret_access_key='Vu3H6nQL+TKx66YtBCmLcWE+sC158DaasZREu/8w'
    )

    bucket_name = 'wenchuanyouself'  # Your S3 bucket name

    ### get RSN number
    # df = pd.read_csv(os.getcwd() + '/Outputs/selected_motions.csv')
    # RSN_numbers = (df.loc[:, 'RSN']).values
    # RSN_number = ''
    # for i in RSN_numbers:
    #     RSN_number = RSN_number + str(i) + ','

    ## download_directory for AWS
    # download_directory = os.getcwd() + '/WenChuan_Outputs'
    download_directory = os.path.join(os.getcwd(), 'Outputs')

    df = pd.read_csv(os.getcwd() + '/Outputs/selected_motions.csv')
    RSN_numbers = (df.loc[:, 'RSN']).values
    RSN_number = ''
    for i in RSN_numbers:
        RSN_number = RSN_number + str(i) + ','

    # rsn_list = ['RSN4577', 'RSN4579']
    rsn_list = convert_to_rsn_list(RSN_number)
    print(rsn_list)
    # print(download_directory)

    csv_file = os.getcwd() + '/Datasets/s3_rsn_paths_at.csv'
    csv_file2 = os.getcwd() + '/Datasets/s3_rsn_paths_vt.csv'
    csv_file3 = os.getcwd() + '/Datasets/s3_rsn_paths_dt.csv'
    csv_file4 = os.getcwd() + '/Datasets/s3_rsn_paths_NGAWest2_Wenchuan_AT2files.csv'



    process_rsn_list(rsn_list, csv_file, download_directory)
    process_rsn_list(rsn_list, csv_file2, download_directory)
    process_rsn_list(rsn_list, csv_file3, download_directory)
    process_rsn_list(rsn_list, csv_file4, download_directory)
    # process_rsn_list(rsn_list, csv_file, download_directory)
